from functools import partial
from munch import Munch

from infinishell.infishell import commandgroup, command, get_active_shell
from infinishell.infishell.types import Str, YesNo, Int
from infinishell.infishell.exceptions import BadArguments,BadValue
from infinishell.infishell.utils import strlist

from .filesystem import FilesystemCommands, FilesystemCommandsMixin
from .types import (
    ShareName,
    Share,
    Filesystem,
    OfflineCaching,
    SmbPermissions,
    NetworkSpace
    )
from .options import QueryOptions
from . import (
    EntityCommands,
    roles,
    render_collection,
    render,
    LazyEntityMapping,
    will_render_any_field,
    )
from ..features import feature_command

class FilesystemShareCommandsBase(EntityCommands):

    def _get_lazy_fs_mapping(self, api, ids=None, preloaded=None):
        filter = None if ids is None else api.create_filter().member_of("id", *ids)
        return LazyEntityMapping(Filesystem, api, fields=("id", "name", "type"), filter=filter, preloaded_entities=preloaded)

@commandgroup(parent=FilesystemCommands, prefix="share")
class FilesystemShareCommands(FilesystemCommandsMixin, FilesystemShareCommandsBase):
    ENTITY_TYPE = Share
    QUERY_FIELDS = strlist("""
        name
        filesystem
        inner_path
        enabled
        offline_caching
        """)
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist("""
        description
        """)
    QUERY_SUBSTITUTES = dict(
        inner_path="internal_path",
        )
    SORTABLE_FIELDS = QUERY_FIELDS - strlist("""
        offline_caching
        filesystem
        """)
        

    @feature_command(
        feature='nas',
        status=True,
        description="List existing SMB shares",
        types=(Filesystem.List, YesNo, Share),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Share,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, fs=None, enabled=None, share=None):
        filesystem_by_id = self._get_lazy_fs_mapping(_api, preloaded=fs) if fs else None
        return self._query(_api,
            enabled=enabled,
            filesystem_id=fs,
            id=share,
            _hook=partial(self._query_hook,
                api=_api,
                filesystem_by_id=filesystem_by_id,
                with_fs_info=will_render_any_field("filesystem", "filesystem_type"),
                ),
            )

    def _query_hook(self, shares, api=None, filesystem_by_id=None, with_fs_info=False):
        if with_fs_info:
            if not filesystem_by_id and api:
                filesystem_by_id = self._get_lazy_fs_mapping(api, ids=(e.filesystem_id for e in shares))
            if filesystem_by_id:
                for s in shares:
                    fs = filesystem_by_id[s.filesystem_id]
                    s.update(
                        filesystem=fs.name,
                        filesystem_type=self._get_fs_type(fs),
                        permissions_count=len(s.permissions),
                        )
        return shares


    @feature_command(
        feature='nas',
        status=True,
        description="Create a new SMB share",
        types=(ShareName, Filesystem, Str, OfflineCaching, Str),
        substitute=dict(
            permissions_export="copy_permissions_from",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, fs, internal_path=None, offline_caching=OfflineCaching.MAPPING['MANUAL'], description=None):
        data = Munch(
            name=name,
            filesystem_id=fs,
            inner_path=internal_path,
            offline_caching=offline_caching,
            description=description
        )
        result = self._create(_api, **data)
        print(result)
        if not any(NetworkSpace.get_all(_api, filter=_api.create_filter().eq("service","NAS_SERVICE"))):
            get_active_shell().warning("NAS is not configured - please create a network space to access the export (use config.net_space.create)")

    @feature_command(
        feature='nas',
        status=True,
        description="Enable SMB share",
        types=(Share,),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def enable(self, _api, share):
        if share.enabled:
            return self._result_message(share, "is already enabled")
        self._update(_api, share, enabled=True)
        return self._result_message(share, "enabled")

    @feature_command(
        feature='nas',
        status=True,
        description="Disable SMB share",
        types=(Share,),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def disable(self, _api, share):
        if not share.enabled:
            return self._result_message(share, "is already disabled")
        self._update(_api, share, enabled=False)
        return self._result_message(share, "disabled")

    @feature_command(
        feature='nas',
        status=True,
        description="Change the offline caching hint of an SMB share",
        types=(Share, OfflineCaching),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_offline_caching(self, _api, share, offline_caching):
        self._update(_api, share, offline_caching=offline_caching)

    @feature_command(
        feature='nas',
        status=True,
        description="Delete SMB share",
        types=(Share,)
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, share):
        return self._delete(_api, share)

    @feature_command(
        feature='nas',
        status=True,
        description="Change the description of an SMB share",
        types=(Share, Str,),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_description(self, _api, share, description):
        self._update(_api, share, description=description)
        return self._result_message(share, "Description updated")

@commandgroup(parent=FilesystemShareCommands, prefix="permission")
class PermissionCommands(FilesystemShareCommandsBase, EntityCommands):
    ENTITY_TYPE = Share

    @command(
        description="Add a new SLP entry to an existing SMB share",
        types=(Share, Str, SmbPermissions),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN.POOLADMIN
    def add(self, _api, share, principal, permission):
        newPermission = {'sid': principal, 'access': permission}
        share.permissions.append(newPermission)
        _api.post("shares/{}/permissions".format(share.id), newPermission)
        return ("""Principal "{}" added to share "{}" (permission: {})""".format(principal, share.name, permission))

    @command(
        description="Modify an SLP entry to an existing SMB share",
        types=(Share, Str, SmbPermissions),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN.POOLADMIN
    def modify(self, _api, share, principal, permission):
        for perm in share.permissions:
            if perm.sid == principal:
                perm_access = {"access":permission}
                perm_id = perm.id
                _api.put("shares/{}/permissions/{}".format(share.id, perm_id), perm_access) 
                return ('Principal "{}" modified on share "{}" (permission: {})'.format(principal, share.name, permission))
        raise BadValue('Principal "{}" is not defined for share "{}"'.format(principal, share.name))

    @command(
        description="Remove an SLP entry to an existing SMB share",
        types=(Share, Str),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN.POOLADMIN
    def remove(self, _api, share, principal):
        for idx, perm in enumerate(share.permissions):
            if perm.sid == principal:
                perm_id = perm.id
                _api.delete("shares/{}/permissions/{}".format(share.id, perm_id))  
                return ('Principal "{}" removed from share "{}" (permission: {})'.format(principal, share.name, perm.access))
        raise BadValue('Principal "{}" is not defined for share "{}"'.format(principal, share.name))

    QUERY_FIELDS = strlist("name filesystem principal permissions")
    SORTABLE_FIELDS = strlist("name")
    QUERY_SUBSTITUTES = dict(
        name="share"
        )

    @feature_command(
        feature='nas',
        status=True,
        description="List SLP entries to SMB shares",
        types=(Share, Filesystem.List),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render(
        fields=QUERY_FIELDS,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, share=None, fs=None):
        filesystem_by_id = self._get_lazy_fs_mapping(_api, preloaded=fs) if fs else None
        return self._query(_api,
            id=share,
            filesystem_id=fs,
            _hook=partial(self._query_hook,
                api=_api,
                filesystem_by_id=filesystem_by_id,
                ),
            )

    def _query_hook(self, shares, api=None, filesystem_by_id=None):
        if not filesystem_by_id and api:
                filesystem_by_id = self._get_lazy_fs_mapping(api, ids=(e.filesystem_id for e in shares))
        if filesystem_by_id:
                for s in shares:
                    fs = filesystem_by_id[s.filesystem_id]
                    s.update(
                        filesystem=fs.name,
                        permissions=[permission.access for permission in s.permissions],
                        principal=[permission.sid for permission in s.permissions]
                        )
        return shares
    
