from functools import partial
from infinishell.infishell import commandgroup, command
from infinishell.infishell.utils import strlist
from . import render_collection, LazyEntityMapping, LazyIdNameMapping, EntityCommands, roles, set_rendered_unit
from .types import Vvol, VVolType, VvolPool, Vm, Name, CapacityUnit, YesNo
from .options import QueryOptions
from .utils import format_savings
from ..features import feature_command

QUERY_FIELDS = strlist("vm vvol_type name type thin size used data_reduction allocated tree_allocated pool_name")
DETAILED_QUERY_FIELDS = strlist("vm vvol_type name type thin size used data_reduction allocated tree_allocated pool_name created_at updated_at ")
SORTABLE_FIELDS = strlist("name type vvol_type created_at updated_at size")
SUBSTITUTE = dict(pool_name="pool", tree_allocated="snapshots")

VM_QUERY_FIELDS = strlist("id name")
VM_SORTABLE_FIELDS = strlist("id name")
VM_SUBSTITUTE = dict(id='vmid',)

@commandgroup(prefix="vvol")
class VvolCommands(EntityCommands):
    ENTITY_TYPE = Vvol

    @feature_command(
        feature='vvol',
        status=True,
        description="List existing vVols",
        types=(Vvol.List, VvolPool, VVolType, Vm, CapacityUnit, YesNo),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=SUBSTITUTE, detailed=True),
        substitute=SUBSTITUTE,
    )
    @render_collection(
        entity=Vvol,
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        detailed_individually=True,
        substitute=SUBSTITUTE,
        alignment=dict(
            data_reduction="right",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN.READONLY
    def query(self, _api, vvol=None, pool_name=None, vvol_type=None, vm=None, unit=None, thin=None):
        set_rendered_unit(unit)
        vm_names_by_id = LazyIdNameMapping(Vm, _api)
        prov_type = None
        if thin is not None:
            prov_type = "THIN" if thin else "THICK"
        return self._query(_api, vvol, pool_id=pool_name, vvol_type=vvol_type, vm_id=vm, prov_type=prov_type,
         _hook=partial(self._query_hook, vm_names_by_id=vm_names_by_id))

    def _query_hook(self, vvols, vm_names_by_id):
        for vvol in vvols:
            vvol.vm = vm_names_by_id[vvol.vm_id]
            vvol.thin = vvol.prov_type == "THIN"
            vvol.data_reduction = format_savings(vvol) 
            yield vvol

    @feature_command(
        feature='vvol',
        status=True,
        description="Delete a vVol",
        types=(Vvol,),
        hidden_substitue=dict(name="vvol"),
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, vvol):
        _api.delete(u"vvols/{}".format(vvol.id))

@commandgroup(prefix="vm", parent=VvolCommands)
class VmCommands(EntityCommands):
    ENTITY_TYPE = Vm

    @feature_command(
        feature='vvol',
        status=True,
        description="List existing VMs",
        options=QueryOptions(fields=VM_QUERY_FIELDS, sortable=VM_QUERY_FIELDS),
        substitute=VM_SUBSTITUTE,
    )
    @render_collection(
        entity=Vm,
        fields=VM_QUERY_FIELDS,
        substitute=VM_SUBSTITUTE,
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN.READONLY
    def query(self, _api):
        return self._query(_api, _hook=self._query_hook)

    def _query_hook(self, vvols):
        for vvol in vvols:
            yield vvol

    @feature_command(
        feature='vvol',
        status=True,
        description="Rename VM",
        types=(Vm, Name),
        hidden_substitue=dict(name="vm"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def rename(self, _api, vm, new_name):
        return self._rename(_api, vm, new_name)
