from munch import Munch

from infinishell.infishell import availability, command, config
from .runtime_config import RuntimeConfig
from infinishell.infishell.utils import strlist

from ..utils import Version
from .system import system, Roles
from infinishell.infishell.exceptions import CommandFailed


runtimeConfig = RuntimeConfig.get_instance()

class Features(object):

    def _route_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.network_space.route_operations_enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _openldap_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.ldap.openldap_enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _openvms_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.openvms.enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _qos_fs_commands(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['mgmt'].get(
                'mgmt.qos.filesystem_support_enabled', feature.get('enabled', False))
        except:
            return feature.get('enabled', False)

    def _smb_commands(self, feature, system):
        try:
            plugins = system.api.get('plugins')
            smb = next(x for x in plugins if x.type == 'SMB')
            return smb is not None
        except:
            return feature.get('enabled', False)

    def _iguard_commands(self, feature, system):
        try:
            integrations = system.api.get('integrations/config')
            return integrations.general.system_type == 'InfiniGuard'
        except:
            return feature.get('enabled', False)
    
    def _fips_commands(self, feature, system):
        try:
            security_mode = system.api.get('config/security/system_security_mode', None)
            return security_mode != "STANDARD"
        except:
            return feature.get('enabled', False)

    def _nas_enabled(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['core'].get(
                'nas.enabled', feature.get('enabled', True))
        except:
            return feature.get('enabled', True)

    def nas_ssd_cache_enabled(self, feature, system):
        try:
            return runtimeConfig['mgmt_config']['core'].get(
                'nas.ssd_cache_enabled', feature.get('enabled', True))
        except:
            return feature.get('enabled', True)

    #############################
    # enabled = whether the feature is enabled by default
    # persistent = whether feature was already tested with test_func/received
    # from server
    # test_func = test_func to see if feature should be enabled
    #############################
    _features = Munch(
        sync_replication=dict(enabled=False, persistant=False, test_func=None),
        qos=dict(enabled=True, persistant=False, test_func=None),
        iscsi=dict(enabled=True, persistant=False, test_func=None),
        treeq=dict(enabled=True, persistant=False, test_func=None),
        vvol=dict(enabled=False, persistant=False, test_func=None),
        vvol_replication=dict(enabled=True, persistant=False, test_func=None),
    )

    _mgmt_features = Munch(
        qos_fs=dict(enabled=False, persistant=False, test_func=_qos_fs_commands),
        route=dict(enabled=True, persistant=False, test_func=_route_commands),
        openldap=dict(enabled=True, persistant=False, test_func=_openldap_commands),
        openvms=dict(enabled=False, persistant=False, test_func=_openvms_commands),
        smb=dict(enabled=False, persistant=False, test_func=_smb_commands),
        iguard=dict(enabled=False, persistant=False, test_func=_iguard_commands),
        fips=dict(enabled=False, persistant=False, test_func=_fips_commands),
        nas=dict(enabled=False, persistant=False, test_func=_nas_enabled),
        ssdCache=dict(enabled=False, persistant=False, test_func=nas_ssd_cache_enabled),
    )

    @classmethod
    def set_mgmt_features(self, system):
        for name, feature in self._mgmt_features.items():
            self.set_feature(name, feature, system)

    @classmethod
    def set_feature(self, feature, info, system):
        self._features.setdefault(feature, dict())
        self._features[feature].update(info)
        test_func = self._features[feature].get('test_func', None)
        persistant = self._features[feature].get('persistant', False)
        if test_func and not persistant:
            self._features[feature] = info
            self._features[feature]['enabled'] = test_func(self, feature=info, system=system)
        elif not persistant:
            self._features[feature] = info
        self._features[feature]['persistant'] = True

    @classmethod
    def get_feature(self, feature_name):
        feature = self._features.get(feature_name, dict(enabled=False, persistant=False, test_func=None))
        assert feature , "Feature {} does not exist in _features list, but is used in a command".format(feature_name)
        test_func = feature.get('test_func', None)
        persistant = feature.get('persistant', False)
        if test_func and not persistant:
            feature['enabled'] = test_func(self, feature=feature, system=system)
            feature['persistant'] = True
        return feature


def feature_command(**kwargs):
    def decorator(_func):
        feature = kwargs.pop('feature')
        status = kwargs.pop('status')
        return availability(command(_func, **kwargs), predicate=lambda: feature_exists(feature) == status)
    def matrix_decorator(_func):
        matrix = kwargs.pop('matrix')
        return availability(command(_func, **kwargs), predicate=lambda: feature_matrix(matrix))
    if kwargs.get('feature') is None or not isinstance(kwargs.get('status'), (bool)):
        if kwargs.get('matrix') is None:
            raise CommandFailed("feature_command decorator requires feature and status parameters or matrix parameter")
        return matrix_decorator
    return decorator

def feature_exists(feature=None):
    return Features.get_feature(feature).get('enabled', False)

# This helper function is for use as optional predicate in commandgroup decorator,
# as e.g. @commandgroup(..., predicate=feature_status("nas", True)) or
# as e.g. @commandgroup(..., predicate=feature_status(matrix=[{"nas": True, "smb": False}]))
# if `matrix` parameter is used (see below), `feature` and `enabled` parameters are ignored.
# if predicate on commandgroup is specified but returns false, all commands of
# that group are disabled, but not subgroups - if those exist and have active commands themselves.
def feature_status(feature=None, enabled=False, matrix=None):
    def predicate():
        return feature_exists(feature)==enabled
    def matrix_predicate():
        return feature_matrix(matrix)
    if matrix is not None:
        return matrix_predicate
    return predicate

# This helper function expects a list of feature dictionaries, each of the kind
# { "feature_name": True/False, ...} to define conditions necessary for command
# availability. For command/commandgroup to be available, at least one of the
# dictionaries has to fit the feature states of the target machine
#
# EXAMPLE:
# feature_option_list = [{"iguard": False, "nas": True}, {"iguard": False, "smb": True}]
# will enable command only if "iguard" feature is absent, and either "nas" or "smb" are present
#
# feature_option_list is called "matrix" in decorators and feature status helper
def feature_matrix(feature_option_list=[]):
    return any([
        all([feature_exists(feature)==enabled
            for (feature, enabled) in feature_option.items()])
        for feature_option in feature_option_list])
